# This file is auto-generated. Do not edit

import functools
from ovs_dbg import decoders

field_decoders = {
    "dp_hash": decoders.Mask32,
    "recirc_id": decoders.decode_int,
    "packet_type": decoders.decode_default,
    "conj_id": decoders.decode_int,
    "tun_id": decoders.Mask64,
    "tunnel_id": decoders.Mask64,
    "tun_src": decoders.IPMask,
    "tun_dst": decoders.IPMask,
    "tun_ipv6_src": decoders.IPMask,
    "tun_ipv6_dst": decoders.IPMask,
    "tun_flags": decoders.decode_default,
    "tun_ttl": decoders.decode_int,
    "tun_tos": decoders.decode_int,
    "tun_gbp_id": decoders.Mask16,
    "tun_gbp_flags": decoders.Mask8,
    "tun_erspan_idx": decoders.decode_mask(20),
    "tun_erspan_ver": decoders.decode_mask(4),
    "tun_erspan_dir": decoders.decode_mask(1),
    "tun_erspan_hwid": decoders.decode_mask(6),
    "tun_gtpu_flags": decoders.Mask8,
    "tun_gtpu_msgtype": decoders.Mask8,
    "tun_metadata0": decoders.Mask992,
    "tun_metadata1": decoders.Mask992,
    "tun_metadata2": decoders.Mask992,
    "tun_metadata3": decoders.Mask992,
    "tun_metadata4": decoders.Mask992,
    "tun_metadata5": decoders.Mask992,
    "tun_metadata6": decoders.Mask992,
    "tun_metadata7": decoders.Mask992,
    "tun_metadata8": decoders.Mask992,
    "tun_metadata9": decoders.Mask992,
    "tun_metadata10": decoders.Mask992,
    "tun_metadata11": decoders.Mask992,
    "tun_metadata12": decoders.Mask992,
    "tun_metadata13": decoders.Mask992,
    "tun_metadata14": decoders.Mask992,
    "tun_metadata15": decoders.Mask992,
    "tun_metadata16": decoders.Mask992,
    "tun_metadata17": decoders.Mask992,
    "tun_metadata18": decoders.Mask992,
    "tun_metadata19": decoders.Mask992,
    "tun_metadata20": decoders.Mask992,
    "tun_metadata21": decoders.Mask992,
    "tun_metadata22": decoders.Mask992,
    "tun_metadata23": decoders.Mask992,
    "tun_metadata24": decoders.Mask992,
    "tun_metadata25": decoders.Mask992,
    "tun_metadata26": decoders.Mask992,
    "tun_metadata27": decoders.Mask992,
    "tun_metadata28": decoders.Mask992,
    "tun_metadata29": decoders.Mask992,
    "tun_metadata30": decoders.Mask992,
    "tun_metadata31": decoders.Mask992,
    "tun_metadata32": decoders.Mask992,
    "tun_metadata33": decoders.Mask992,
    "tun_metadata34": decoders.Mask992,
    "tun_metadata35": decoders.Mask992,
    "tun_metadata36": decoders.Mask992,
    "tun_metadata37": decoders.Mask992,
    "tun_metadata38": decoders.Mask992,
    "tun_metadata39": decoders.Mask992,
    "tun_metadata40": decoders.Mask992,
    "tun_metadata41": decoders.Mask992,
    "tun_metadata42": decoders.Mask992,
    "tun_metadata43": decoders.Mask992,
    "tun_metadata44": decoders.Mask992,
    "tun_metadata45": decoders.Mask992,
    "tun_metadata46": decoders.Mask992,
    "tun_metadata47": decoders.Mask992,
    "tun_metadata48": decoders.Mask992,
    "tun_metadata49": decoders.Mask992,
    "tun_metadata50": decoders.Mask992,
    "tun_metadata51": decoders.Mask992,
    "tun_metadata52": decoders.Mask992,
    "tun_metadata53": decoders.Mask992,
    "tun_metadata54": decoders.Mask992,
    "tun_metadata55": decoders.Mask992,
    "tun_metadata56": decoders.Mask992,
    "tun_metadata57": decoders.Mask992,
    "tun_metadata58": decoders.Mask992,
    "tun_metadata59": decoders.Mask992,
    "tun_metadata60": decoders.Mask992,
    "tun_metadata61": decoders.Mask992,
    "tun_metadata62": decoders.Mask992,
    "tun_metadata63": decoders.Mask992,
    "metadata": decoders.Mask64,
    "in_port": decoders.decode_default,
    "in_port_oxm": decoders.decode_default,
    "actset_output": decoders.decode_default,
    "skb_priority": decoders.decode_int,
    "pkt_mark": decoders.Mask32,
    "ct_state": decoders.decode_default,
    "ct_zone": decoders.decode_int,
    "ct_mark": decoders.Mask32,
    "ct_label": decoders.Mask128,
    "ct_nw_proto": decoders.decode_int,
    "ct_nw_src": decoders.IPMask,
    "ct_nw_dst": decoders.IPMask,
    "ct_ipv6_src": decoders.IPMask,
    "ct_ipv6_dst": decoders.IPMask,
    "ct_tp_src": decoders.Mask16,
    "ct_tp_dst": decoders.Mask16,
    "reg0": decoders.Mask32,
    "reg1": decoders.Mask32,
    "reg2": decoders.Mask32,
    "reg3": decoders.Mask32,
    "reg4": decoders.Mask32,
    "reg5": decoders.Mask32,
    "reg6": decoders.Mask32,
    "reg7": decoders.Mask32,
    "reg8": decoders.Mask32,
    "reg9": decoders.Mask32,
    "reg10": decoders.Mask32,
    "reg11": decoders.Mask32,
    "reg12": decoders.Mask32,
    "reg13": decoders.Mask32,
    "reg14": decoders.Mask32,
    "reg15": decoders.Mask32,
    "xreg0": decoders.Mask64,
    "xreg1": decoders.Mask64,
    "xreg2": decoders.Mask64,
    "xreg3": decoders.Mask64,
    "xreg4": decoders.Mask64,
    "xreg5": decoders.Mask64,
    "xreg6": decoders.Mask64,
    "xreg7": decoders.Mask64,
    "xxreg0": decoders.Mask128,
    "xxreg1": decoders.Mask128,
    "xxreg2": decoders.Mask128,
    "xxreg3": decoders.Mask128,
    "eth_src": decoders.EthMask,
    "dl_src": decoders.EthMask,
    "eth_dst": decoders.EthMask,
    "dl_dst": decoders.EthMask,
    "eth_type": decoders.decode_int,
    "dl_type": decoders.decode_int,
    "vlan_tci": decoders.Mask16,
    "dl_vlan": decoders.decode_int,
    "vlan_vid": decoders.decode_mask(12),
    "dl_vlan_pcp": decoders.decode_int,
    "vlan_pcp": decoders.decode_int,
    "mpls_label": decoders.decode_int,
    "mpls_tc": decoders.decode_int,
    "mpls_bos": decoders.decode_int,
    "mpls_ttl": decoders.decode_int,
    "ip_src": decoders.IPMask,
    "nw_src": decoders.IPMask,
    "ip_dst": decoders.IPMask,
    "nw_dst": decoders.IPMask,
    "ipv6_src": decoders.IPMask,
    "ipv6_dst": decoders.IPMask,
    "ipv6_label": decoders.decode_mask(20),
    "nw_proto": decoders.decode_int,
    "ip_proto": decoders.decode_int,
    "nw_tos": decoders.decode_int,
    "ip_dscp": decoders.decode_int,
    "nw_ecn": decoders.decode_int,
    "ip_ecn": decoders.decode_int,
    "nw_ttl": decoders.decode_int,
    "ip_frag": decoders.decode_default,
    "nw_frag": decoders.decode_default,
    "arp_op": decoders.decode_int,
    "arp_spa": decoders.IPMask,
    "arp_tpa": decoders.IPMask,
    "arp_sha": decoders.EthMask,
    "arp_tha": decoders.EthMask,
    "tcp_src": decoders.Mask16,
    "tp_src": decoders.Mask16,
    "tcp_dst": decoders.Mask16,
    "tp_dst": decoders.Mask16,
    "tcp_flags": decoders.decode_default,
    "udp_src": decoders.Mask16,
    "udp_dst": decoders.Mask16,
    "sctp_src": decoders.Mask16,
    "sctp_dst": decoders.Mask16,
    "icmp_type": decoders.decode_int,
    "icmp_code": decoders.decode_int,
    "icmpv6_type": decoders.decode_int,
    "icmpv6_code": decoders.decode_int,
    "nd_target": decoders.IPMask,
    "nd_sll": decoders.EthMask,
    "nd_tll": decoders.EthMask,
    "nd_reserved": decoders.decode_int,
    "nd_options_type": decoders.decode_int,
    "nsh_flags": decoders.Mask8,
    "nsh_mdtype": decoders.decode_int,
    "nsh_np": decoders.decode_int,
    "nsh_spi": decoders.decode_int,
    "nsp": decoders.decode_int,
    "nsh_si": decoders.decode_int,
    "nsi": decoders.decode_int,
    "nsh_c1": decoders.Mask32,
    "nshc1": decoders.Mask32,
    "nsh_c2": decoders.Mask32,
    "nshc2": decoders.Mask32,
    "nsh_c3": decoders.Mask32,
    "nshc3": decoders.Mask32,
    "nsh_c4": decoders.Mask32,
    "nshc4": decoders.Mask32,
    "nsh_ttl": decoders.decode_int,
}
